// ============================================================================
// decodeCave.cpp
// ----------------------------------------------------------------------------
// (c) 2022-2024 by HerzAusGold
//               Hubert Haberkorn
//               herz-aus-gold-42@gmx.de
//               http://www.arthurdent.de/
// ----------------------------------------------------------------------------
// The above copyright notice and this permission notice 
// shall be included in all copies or substantial portions 
// of the Software.
// ----------------------------------------------------------------------------
// read "raw" file dump and extract the data with parameters.
// create .WLC file - WithLevelCommon
// ----------------------------------------------------------------------------

#include "decodeCave.h"
#include <string>
#include <assert.h>

static UBYTE caveData[MAXX*MAXY];
static UBYTE realCaveData[MAXX*MAXY];

/* DrawLine data */
/* When drawing lines, you can draw in all eight directions. This table gives
   the offsets needed to move in each of the 8 directions. */
static int ldx[8]={ 0,  1, 1, 1, 0, -1, -1, -1};
static int ldy[8]={-1, -1, 0, 1, 1,  1,  0, -1};

/* Human-readable versions of the 8 directions */
static const char* directions[]=
{
    "up","up/right","right","down/right","down","down/left","left","up/left"
};

/* C64 color names */
static const char* colors[]=
{
    "Black", "White", "Red", "Cyan", "Purple", "Green", "Blue", "Yellow", "Orange",
    "Brown", "Light red", "Gray 1", "Gray 2", "Light green", "Light blue", "Gray 3"
};

/* Human-readable names for the 64 different C64 object codes */
static const char* objects[]=
{
    "zSpace",  "zDirt",  "zBrick", "zMagic", "zPreOut", "zOutBox", "*",       "zSteel",
    "zFFly1",  "zFFly2", "zFFly3", "zFFly4", "zFFly1M", "zFFly2M", "zFFly3M", "zFFly4M", 
    "zBouS",   "zBouSM", "zBouF",  "zBouFM", "zDiaS",   "zDiaSM",  "zDiaF",   "zDiaFM",  
    "",        "",       "",       "zExp1S", "zExp2S",  "zExp3S",  "zExp4S",  "zExp5S",
    "zExp1D",  "zExp2D", "zExp3D", "zExp4D", "zExp5D",  "zPRFd1",  "zPRFd2",  "zPRFd3",
    "zPRFd4",  "",       "",       "",       "",        "",        "",        "",
    "zBFly1",  "zBFly2", "zBFly3", "zBFly4", "zBFly1M","zBFly2M", "zBFly3M", "zBFly4M", 
    "zRFd",    "zRFdM",  "zAmoe",  "zAmoeM", "",       "",        "*",       "*"
};

/* Creature code conversion table */
/* Converts the C64 BoulderDash codes into the codes used by Jeff Bevis's
   Amiga implementation of BoulderDash. */
static unsigned char creatureCode[64]=
{ //00   01   02   03   04   05   06   07
    ' ', '.', 'w', 'm', 'P', 'P', '?', 'W',
  //08   09   0A   0B   0C   0D   0E   0F
    'q', 'o', 'Q', 'O', 'q', 'o', 'Q', 'O',
  //10   11   12   13   14   15   16   17
    'r', 'r', 'r', 'r', 'd', 'd', 'd', 'd',
  //18   19   1A   1B   1C   1D   1E   1F
    '?', '?', '?', '?', '?', '?', '?', '?',
  //20   21   22   23   24   25   26   27
    '?', '?', '?', '?', '3', 'X', '5', '6',
  //28   29   2A   2B   2C   2D   2E   2F
    '?', '?', 'E', '?', '?', '?', '?', '?',         // 2A
  //30   31   32   33   34   35   36   37
    'B', 'c', 'b', 'C', 'B', 'c', 'b', 'C',
  //38   39   3A   3B   3C   3D   3E   3F           // 3E
    '7', '8', 'a', 'a', '?', '?', 'S', '?' 
};


/* **************************************** */
char GetCreatorCode(UBYTE ch) {
    char retval= '?';
    /*--
    switch(ch){
        case 0x10    : retval = 'r';  break;  // retval = 'o';  break;  // rock
        case 0x14    : retval = 'd';  break;  // retval = '*';  break;  // diamond
        case 0x03    : retval = 'm';  break;  // retval = 'm';  break;  // magicWall
        case 0x02    : retval = 'w';  break;  // retval = 'W';  break;  // wall
        case 0x07    : retval = 'W';  break;  // retval = '#';  break;  // steelWall
        case 0x2A    : retval = 'E';  break;  // retval = 'e';  break;  // expandWall
        case 0x3C    : retval = 'R';  break;  // retval = 'r';  break;  // rockford (fake like wall)
        case 0x01    : retval = '.';  break;  // retval = ':';  break;  // dirt
        case 0x08    : retval = 'q';  break;  // retval = 'f';  break;  // firefly   (links)
        case 0x30    : retval = 'B';  break;  // retval = 'b';  break;  // butterfly
        case 0x3A    : retval = 'a';  break;  // retval = 'a';  break;  // amoeba
        case 0x3E    : retval = 'S';  break;  // retval = 's';  break;  // slime
        case 0x3D    : retval = 'H';  break;  // retval = 'H';  break;  // hiddenOut
        case 0x04    : retval = 'P';  break;  // retval = 'V';  break;  // visibleOut
        case 0x25    : retval = 'X';  break;  // retval = 'P';  break;  // player
        case 0x00    : retval = ' ';  break;  // retval = ' ';  break;  // space (empty)
        default      : retval = '?';  break;
    }
    --*/
    if ((ch >= 0) && (ch <= 63)) {
        retval = creatureCode[ch];
    } 
    if (retval == '?') {
        retval = '?';    // breakpoint here
    }
    return retval;
}

/* **************************************** */
const char * GetDirections(UBYTE dir)
{
    static const char* unknown = "???";
    if ((dir >= 0) && (dir <= 7)) {
        return directions[dir];
    } else {
        return unknown;
    }
}

/* **************************************** */
const char * GetObjects(UBYTE obj)
{
    static const char* unknown = "?????";
    static char hex[20] = {0};

    if ((obj >= 0) && (obj <= 63)) {
        if (strlen(objects[obj]) == 0) {
            sprintf(hex, "0x%2.2X", obj);
            return hex;
        } else {
            return objects[obj];
        }
    } else {
        return unknown;
    }
}

/* **************************************** */
const char * GetColors(UBYTE col)
{
    static const char* unknown = "???";
    if (col >= 0 && col <= 16) {
        return colors[col];
    } else {
        return unknown;
    }
}

/* **************************************** */
UBYTE GetCaveData(int x, int y) {
    int pos = y*MAXX + x;
    return caveData[pos];
}

/* **************************************** */
UBYTE GetRealCaveData(int x, int y) {
    int pos = y*MAXX + x;
    return realCaveData[pos];
}

/* **************************************** */
UBYTE * GetRealCaveDataPointer(int x, int y) {
    int pos = y*MAXX + x;
    return &realCaveData[pos];
}

/* **************************************** */
void StoreObject(int x, int y, objectType anObject)
{
    //assert(((x >= 0) && (x <= 39)));
    //assert(((y >= 0) && (y <= 23)));
    assert(((anObject >= 0) && (anObject <= 63)));
 
    if ((y*MAXX+x) >= sizeof(caveData)) {
        int pos = y*MAXX+x;  //??
    } else {
        caveData[y*MAXX+x] = creatureCode[anObject];
        realCaveData[y*MAXX+x] = (UBYTE) anObject;
    }
}

/* **************************************** */
void StoreObjectDirect(int addr, objectType anObject)
{
    //assert(((x >= 0) && (x <= 39)));
    //assert(((y >= 0) && (y <= 23)));
    assert(((anObject >= 0) && (anObject <= 63)));
 
    if ((addr) >= sizeof(caveData)) {
        int pos = addr;  //??
    } else {
        caveData[addr] = creatureCode[anObject];
        realCaveData[addr] = (UBYTE) anObject;
    }
}

/* **************************************** */
void DrawLine(objectType anObject, int x, int y, int aLength, int aDirection)
{
/* Local variables */
    int counter;

    //assert(((anObject >= 0) && (anObject <= 63)));
    //assert(((x >= 0) && (x <= 39)));
    //assert(((y >= 0) && (y <= 23)));
    //assert(((aLength >= 1) && (aLength <= 40)));
    //assert(((aDirection >= 0) && (aDirection <= 7)));

    for(counter = 1; counter <= aLength; counter++) {
        StoreObject(x, y, anObject);
        x += ldx[aDirection];
        y += ldy[aDirection];
    }
}

/* **************************************** */
void DrawFilledRectSimple(objectType anObject, int x, int y, int aWidth, int aHeight)
{
/* Local variables */
    int counter1;
    int counter2;

    for(counter1 = 0; counter1 < aWidth; counter1++) {
        for(counter2 = 0; counter2 < aHeight; counter2++) {
            StoreObject(x+counter1, y+counter2, anObject);
        }
    }
}

/* **************************************** */
void DrawFilledRect(objectType anObject, int x, int y, int aWidth, int aHeight, objectType aFillObject)
{
/* Local variables */
//    int counter1;
//    int counter2;

    // assert(((anObject >= 0) && (anObject <= 63)));
    // assert(((x >= 0) && (x <= 39)));
    // assert(((y >= 0) && (y <= 23)));
    // assert(((aWidth >= 1) && (aWidth <= 40)));
    // assert(((aHeight >= 1) && (aHeight <= 24)));
    // assert(((aFillObject >= 0) && (aFillObject <= 63)));

    /*--
    for(counter1 = 0; counter1 < aWidth; counter1++) {
        for(counter2 = 0; counter2 < aHeight; counter2++) {
            if ((counter2 == 0) || (counter2 == aHeight-1)) {
                StoreObject(x+counter1, y+counter2, anObject);
            } else {
                StoreObject(x+counter1, y+counter2, aFillObject);
            }
        }
        StoreObject(x+counter1, y, anObject);
        StoreObject(x+counter1, y+aHeight-1, anObject);
    }
    --*/

    DrawRect(anObject, x, y, aWidth, aHeight);
    x += 1;
    y += 1;
    aWidth  -= 2;
    aHeight -= 2;
    DrawFilledRectSimple(aFillObject, x, y, aWidth, aHeight);
}

/* **************************************** */
void DrawRect(objectType anObject, int x, int y, int aWidth, int aHeight)
{
/* Local variables */
    int counter1;

    // assert(((anObject >= 0) && (anObject <= 63)));
    // assert(((x >= 0) && (x <= 39)));
    // assert(((y >= 0) && (y <= 23)));
    // assert(((aWidth >= 1) && (aWidth <= 40)));
    // assert(((aHeight >= 1) && (aHeight <= 24)));

    for(counter1 = 0; counter1 < aWidth; counter1++) {
        StoreObject(x + counter1, y, anObject);
        StoreObject(x + counter1, y + aHeight-1, anObject);
    }
    for(counter1 = 0; counter1 < aHeight; counter1++) {
        StoreObject(x, y + counter1, anObject);
        StoreObject(x + aWidth-1, y + counter1, anObject);
    }
}

/* **************************************** */
void NextRandom(int *RandSeed1, int *RandSeed2)
/* This is the mathematical random number generator from the Commodore 64
   implementation of Boulder Dash I. The 6510 disassembly is given in
   comments, and the C translation follows. */
{
    short TempRand1;
    short TempRand2;
    short carry;
    short result;

    assert(((*RandSeed1 >= 0x00) && (*RandSeed1 <= 0xFF)));
    assert(((*RandSeed2 >= 0x00) && (*RandSeed2 <= 0xFF)));

/*
            7085 NextRandom.00  LDA RandSeed1
            7087                ROR
            7088                ROR
            7089                AND #$80
            708B                STA TempRand1
        
            Note: ROR on the 6510 works like this:
              7-->6-->5-->4-->3-->2-->1-->0-->C
              ^                               |
              |_______________________________|
            In other words, it's a nine-bit rotate. Thus it takes two RORs to shift
            the low bit (bit zero) into the high bit (bit 7).
*/
    TempRand1 = (*RandSeed1 & 0x0001) * 0x0080;	/* Bugfix! */


        /*
            708E                LDA RandSeed2
            7090                ROR
            7091                AND #$7F
            7093                STA TempRand2
        */
    TempRand2 = (*RandSeed2 >> 1) & 0x007F;


        /*
            7096                LDA RandSeed2
            7098                ROR
            7099                ROR
            709A                AND #$80
            709C                CLC
            709D                ADC RandSeed2
        */
    result = (*RandSeed2) + (*RandSeed2 & 0x0001) * 0x0080;
    carry = (result > 0x00FF);
    result = result & 0x00FF;


        /*
            709F                ADC #$13
            70A1                STA RandSeed2
        */
    result = result + carry + 0x13;
    carry = (result > 0x00FF);
    *RandSeed2 = result & 0x00FF;


        /*
            70A3                LDA RandSeed1
            70A5                ADC TempRand1
        */
    result = *RandSeed1 + carry + TempRand1;
    carry = (result > 0x00FF);
    result = result & 0x00FF;


        /*
            70A8                ADC TempRand2
            70AB                STA RandSeed1
            70AD                RTS
        */
    result = result + carry + TempRand2;
    *RandSeed1 = result & 0x00FF;

    assert(((*RandSeed1 >= 0x00) && (*RandSeed1 <= 0xFF)));
    assert(((*RandSeed2 >= 0x00) && (*RandSeed2 <= 0xFF)));
}

/* **************************************** */
/* **************************************** */

